﻿Imports System.Text
Imports System.IO

Public Class WaveHeader

    Public Const RiffID = "RIFF"

    'RIFF_HEADER
    Public RiffSize As UInt32 'little_endian
    Public Const RiffFormat = "WAVE"

    'FMT_BLOCK
    Public Const FmtID = "fmt "
    Public FmtSize As UInt32 'little_endian

    'WAVE_FORMAT
    '音频数据的编码方式 1 表示是 PCM 编码, 声道数
    Public FormatTag, Channels As UInt16
    '采样率(每秒样本数), 音频数据传送速率, 单位是字节
    Public SamplesPerSec, AvgBytesPerSec As UInt32
    '每次采样的大小 = 采样精度*声道数/8(单位是字节),  每个声道的采样精度
    Public BlockAlign, BitsPerSample As UInt16

    'FACT_BLOCK
    Public HasFact As Boolean = False
    Public Const FactID = "fact"
    Public FactSize As UInt32 'little_endian
    Public FactData(4) As Byte

    'DATA_BLOCK
    Public Const DataID = "data"
    Public DataSize As UInt32 'little_endian
    'skip(DataSize);

    Public Function Length() As Integer
        Dim riff_fmt = 4 + 4 + 4 + 4 + 4 + FmtSize
        Dim fact = 4 + 4 + FactSize
        Dim data = 4 + 4 '+ DataSize
        Dim len = riff_fmt + data
        If HasFact Then
            len += fact
        End If
        Return len
    End Function

    Public Shared Function ReadString(s As Stream, length As Integer) As String
        Dim data(length - 1) As Byte
        Dim read = 0
        Do While read < length
            read += s.Read(data, read, length - read)
        Loop
        Return Encoding.ASCII.GetString(data)
    End Function

    Public Shared Function ReadUInt16LE(s As Stream) As UInt16
        Dim data As UInt16 = 0
        data += s.ReadByte() << 0
        data += s.ReadByte() << 8
        Return data
    End Function

    Public Shared Function ReadUInt32LE(s As Stream) As UInt32
        Dim data As UInt32 = 0
        data += s.ReadByte() << 0
        data += s.ReadByte() << 8
        data += s.ReadByte() << 16
        data += s.ReadByte() << 24
        Return data
    End Function

    Public Shared Sub Skip(s As Stream, length As Integer)
        Try
            s.Seek(length, SeekOrigin.Current)
        Catch ex As Exception
            Dim tmp(4096) As Byte
            Dim read = 0
            Do While read < length
                read += s.Read(tmp, 0, length - read)
            Loop
        End Try
    End Sub

    Public Shared Function Read(s As Stream) As WaveHeader
        Dim wh As New WaveHeader
        If ReadString(s, 4) <> RiffID Then
            Throw New Exception("format mismatch")
        End If
        wh.RiffSize = ReadUInt32LE(s)
        If ReadString(s, 4) <> RiffFormat Then
            Throw New Exception("format mismatch")
        End If

        If ReadString(s, 4) <> FmtID Then
            Throw New Exception("format mismatch")
        End If

        wh.FmtSize = ReadUInt32LE(s)
        wh.FormatTag = ReadUInt16LE(s)
        wh.Channels = ReadUInt16LE(s)
        wh.SamplesPerSec = ReadUInt32LE(s)
        wh.AvgBytesPerSec = ReadUInt32LE(s)
        wh.BlockAlign = ReadUInt16LE(s)
        wh.BitsPerSample = ReadUInt16LE(s)
        Skip(s, wh.FmtSize - 16)

        Dim UnknownID = ReadString(s, 4)

        If UnknownID = FactID Then
            wh.HasFact = True
            wh.FactSize = ReadUInt32LE(s)

            Dim read_len = 0
            Do While read_len < 4
                read_len += s.Read(wh.FactData, read_len, 4 - read_len)
            Loop

            Skip(s, wh.FactSize - 4)

            UnknownID = ReadString(s, 4)
        End If

        If UnknownID = DataID Then
            wh.DataSize = ReadUInt32LE(s)
        End If

        Return wh
    End Function

    Public Shared Sub WriteString(s As Stream, str As String)
        Dim data() = Encoding.Default.GetBytes(str)
        s.Write(data, 0, data.Length)
    End Sub

    Public Shared Sub WriteUInt16LE(s As Stream, i As UInt16)
        s.WriteByte((i >> 0) And 255)
        s.WriteByte((i >> 8) And 255)
    End Sub

    Public Shared Sub WriteUInt32LE(s As Stream, i As UInt32)
        s.WriteByte((i >> 0) And 255)
        s.WriteByte((i >> 8) And 255)
        s.WriteByte((i >> 16) And 255)
        s.WriteByte((i >> 24) And 255)
    End Sub

    Public Shared Sub Write(s As Stream, wh As WaveHeader)
        WriteString(s, RiffID)
        WriteUInt32LE(s, wh.RiffSize)
        WriteString(s, RiffFormat)

        WriteString(s, FmtID)
        WriteUInt32LE(s, wh.FmtSize)
        WriteUInt16LE(s, wh.FormatTag)
        WriteUInt16LE(s, wh.Channels)
        WriteUInt32LE(s, wh.SamplesPerSec)
        WriteUInt32LE(s, wh.AvgBytesPerSec)
        WriteUInt16LE(s, wh.BlockAlign)
        WriteUInt16LE(s, wh.BitsPerSample)

        If wh.FmtSize > 16 Then
            WriteString(s, New String(" "c, wh.FmtSize - 16))
        End If

        If wh.HasFact Then
            WriteString(s, FactID)
            WriteUInt32LE(s, wh.FactSize)
            s.Write(wh.FactData, 0, wh.FactData.Length)
            If wh.FactSize > wh.FactData.Length Then
                WriteString(s, New String(" "c, wh.FactSize - wh.FactData.Length))
            End If
        End If

        WriteString(s, DataID)
        WriteUInt32LE(s, wh.DataSize)

        s.Flush()
    End Sub

    Public Sub Save(s As Stream)
        Write(s, Me)
    End Sub

    Public coding As New Dictionary(Of UInt16, String) From {
        {0, "Unknown"},
        {1, "PCM"},
        {2, "Microsoft ADPCM"},
        {3, "IEEE Float"},
        {6, "ITU G.711 a-law"},
        {7, "ITU G.711 Âµ-law"},
        {17, "IMA ADPCM"},
        {20, "ITU G.723 ADPCM (Yamaha)"},
        {49, "GSM 6.10"},
        {64, "ITU G.721 ADPCM"},
        {80, "MPEG"},
          {65534, "ExtPCM"},
        {65535, "Experimental"}}

    Public Overrides Function ToString() As String

        Dim codstr = ""
        If coding.ContainsKey(Me.FormatTag) Then
            codstr = coding(Me.FormatTag)
        Else
            codstr = "Unknown(" & Me.FormatTag.ToString() & ")"
        End If


        Return String.Format("{0}Hz, {1}bit, {2}, {3} Encoding",
                             Me.SamplesPerSec,
                             Me.BitsPerSample,
                             IIf(Me.Channels > 1, "Stereo", "Mono"),
                             codstr)
    End Function

End Class
