﻿Imports System.IO

Public Class WaveWriter

    Private wroteheader = False
    Public Header As New WaveHeader
    Private _s As Stream

    Sub New(s As Stream, SamplesPerSec As UInt32)
        _s = s
        Header.SamplesPerSec = SamplesPerSec
    End Sub

    Public Sub WriteHeader()
        UpdateHeader()
        If wroteheader Then
            Dim curr = _s.Position
            _s.Seek(0, SeekOrigin.Begin)
            Header.Save(_s)
            wroteheader = True
            _s.Seek(curr, SeekOrigin.Begin)
        Else
            Header.Save(_s)
            wroteheader = True
        End If
    End Sub

    Public Sub UpdateHeader()
        Header.BlockAlign = Header.Channels * Header.BitsPerSample / 8
        Header.AvgBytesPerSec = Header.SamplesPerSec * Header.BlockAlign
        Header.FmtSize = 4 * 2 + 2 * 4
        Header.RiffSize = Header.Length() - 8 + Header.DataSize
    End Sub

    Public Sub GenerateMonoSample(Of T)(time As Double, phase0 As Double, generator0 As Func(Of Double, T))
        setType(GetType(T))
        setChannels(1)
        Dim samplescount = time * Header.SamplesPerSec
        addSamples(samplescount)
        WriteHeader()
        For i = 0 To samplescount - 1
            Dim data0 As T = generator0(i / Header.SamplesPerSec + phase0)
            WriteData(Of T)(_s, data0)
        Next
    End Sub

    Private Shared Sub WriteData(Of T)(s As Stream, data As T)
        Dim data0 As Object = data
        If GetType(T) Is GetType(Byte) Then
            s.WriteByte(CByte(data0))
        ElseIf GetType(T) Is GetType(Int16) Then
            WaveHeader.WriteUInt16LE(s, CUShort(data0 And &HFFFF))
        ElseIf GetType(T) Is GetType(Int32) Then '24 bits
            WaveHeader.WriteUInt16LE(s, CUShort(data0 And &HFFFF))
            s.WriteByte(CByte(data0 >> 16))
        ElseIf GetType(T) Is GetType(Single) Then
            Dim b = BinaryCast.SingleToByte(data0)
            s.Write(b, 0, b.Length)
        Else
            Throw New NotImplementedException(GetType(T).Name)
        End If
    End Sub

    Public Sub GenerateStereoSample(Of T)(time As Double, phase0 As Double, generator0 As Func(Of Double, T), phase1 As Double, generator1 As Func(Of Double, T))
        setType(GetType(T))
        setChannels(2)
        Dim samplescount = time * Header.SamplesPerSec
        addSamples(samplescount)
        WriteHeader()
        For i = 0 To samplescount - 1
            Dim data0 As T = generator0(i / Header.SamplesPerSec + phase0)
            WriteData(Of T)(_s, data0)
            Dim data1 As T = generator1(i / Header.SamplesPerSec + phase1)
            WriteData(Of T)(_s, data1)
        Next
    End Sub

    Public Sub Generate5_1Sample(Of T)(time As Double, phase0 As Double, generator0 As Func(Of Double, T), phase1 As Double, generator1 As Func(Of Double, T), phase2 As Double, generator2 As Func(Of Double, T), phase3 As Double, generator3 As Func(Of Double, T), phase4 As Double, generator4 As Func(Of Double, T), phase5 As Double, generator5 As Func(Of Double, T))
        setType(GetType(T))
        setChannels(6)
        Dim samplescount = time * Header.SamplesPerSec
        addSamples(samplescount)
        WriteHeader()
        For i = 0 To samplescount - 1
            Dim data0 As T = generator0(i / Header.SamplesPerSec + phase0)
            WriteData(Of T)(_s, data0)
            Dim data1 As T = generator1(i / Header.SamplesPerSec + phase1)
            WriteData(Of T)(_s, data1)
            Dim data2 As T = generator2(i / Header.SamplesPerSec + phase2)
            WriteData(Of T)(_s, data2)
            Dim data3 As T = generator3(i / Header.SamplesPerSec + phase3)
            WriteData(Of T)(_s, data3)
            Dim data4 As T = generator4(i / Header.SamplesPerSec + phase4)
            WriteData(Of T)(_s, data4)
            Dim data5 As T = generator5(i / Header.SamplesPerSec + phase5)
            WriteData(Of T)(_s, data5)
        Next
    End Sub

    Public Sub WriteMonoSample(Of T)(Channel0 As T)
        setType(GetType(T))
        setChannels(1)
        addSamples(1)
        'WriteHeader()
        WriteData(Of T)(_s, Channel0)
    End Sub

    Public Sub WriteStereoSample(Of T)(Channel0 As T, Channel1 As T)
        setType(GetType(T))
        setChannels(2)
        addSamples(1)
        'WriteHeader()
        WriteData(Of T)(_s, Channel0)
        WriteData(Of T)(_s, Channel1)
    End Sub

    Public Sub Write5_1Sample(Of T)(Channel0 As T, Channel1 As T, Channel2 As T, Channel3 As T, Channel4 As T, Channel5 As T)
        setType(GetType(T))
        setChannels(6)
        addSamples(1)
        'WriteHeader()
        WriteData(Of T)(_s, Channel0)
        WriteData(Of T)(_s, Channel1)
        WriteData(Of T)(_s, Channel2)
        WriteData(Of T)(_s, Channel3)
        WriteData(Of T)(_s, Channel4)
        WriteData(Of T)(_s, Channel5)
    End Sub

    Public Sub WriteMonoSample(Of T)(Channel0 As T())
        setType(GetType(T))
        setChannels(1)
        addSamples(Channel0.Length)
        WriteHeader()
        For i = 0 To Channel0.Length - 1
            WriteData(Of T)(_s, Channel0(i))
        Next
    End Sub

    Public Sub WriteStereoSample(Of T)(Channel0 As T(), Channel1 As T())
        setType(GetType(T))
        setChannels(2)
        addSamples(Channel0.Length)
        WriteHeader()
        If Channel0.Length <> Channel1.Length Then
            Throw New ArgumentException("Channel0, Channel1 Length mismatch")
        End If
        For i = 0 To Channel0.Length - 1
            WriteData(Of T)(_s, Channel0(i))
            WriteData(Of T)(_s, Channel1(i))
        Next
    End Sub

    Public Sub Write5_1Sample(Of T)(Channel0 As T(), Channel1 As T(), Channel2 As T(), Channel3 As T(), Channel4 As T(), Channel5 As T())
        setType(GetType(T))
        setChannels(6)
        addSamples(Channel0.Length)
        WriteHeader()
        If Not (Channel0.Length = Channel1.Length AndAlso
               Channel1.Length = Channel2.Length AndAlso
               Channel2.Length = Channel3.Length AndAlso
               Channel3.Length = Channel4.Length AndAlso
               Channel4.Length = Channel5.Length) Then
            Throw New ArgumentException("Channel0, Channel1, Channel2, Channel3, Channel4, Channel5 Length mismatch")
        End If
        For i = 0 To Channel0.Length - 1
            WriteData(Of T)(_s, Channel0(i))
            WriteData(Of T)(_s, Channel1(i))
            WriteData(Of T)(_s, Channel2(i))
            WriteData(Of T)(_s, Channel3(i))
            WriteData(Of T)(_s, Channel4(i))
            WriteData(Of T)(_s, Channel5(i))
        Next
    End Sub

    Public Sub addSamples(samplecount As Integer)
        Header.DataSize += Header.BlockAlign * samplecount
        UpdateHeader()
    End Sub

    Public Sub setChannels(Channels As UInt16)
        Header.Channels = Channels
        UpdateHeader()
    End Sub

    Public Sub setType(type As Type)
        If type Is GetType(Byte) Then '8bits PCM
            Header.BitsPerSample = 8
            Header.FormatTag = 1 'PCM
        ElseIf type Is GetType(Int16) Then '16bits PCM
            Header.BitsPerSample = 16
            Header.FormatTag = 1 'PCM
        ElseIf type Is GetType(Int32) Then '24 bits PCM
            Header.BitsPerSample = 24
            Header.FormatTag = 1 'PCM
        ElseIf type Is GetType(Single) Then '32bits IEEE Float
            Header.BitsPerSample = 32
            Header.FormatTag = 3 'IEEE Float
        Else
            Throw New NotImplementedException(type.Name)
        End If
        UpdateHeader()
    End Sub

End Class
